from __future__ import annotations

from functools import reduce
from typing import TYPE_CHECKING, Any, Callable, Final, Generic, TypeVar, overload

from typing_extensions import Concatenate

from ..overloads import SINGLETON_OVERLOAD
from ..typing import CT, P1, Collectable, ImplementSample, P, R
from .record import FnImplement, FnRecord

if TYPE_CHECKING:
    from .base import Fn
    from .implement import OverloadRecorder

FC = TypeVar("FC", bound="FnCompose")


class FnCompose:
    singleton: Final = SINGLETON_OVERLOAD
    fn: Fn

    def __init__(self, fn: Fn):
        self.fn = fn

    def call(self, record: FnRecord, *args, **kwargs) -> Any:
        return next(iter(SINGLETON_OVERLOAD.dig(record, None).keys()))(*args, **kwargs)

    def collect(self, recorder: OverloadRecorder) -> None:
        recorder.use(self.singleton, None)

    def signature(self):
        return FnImplement(self.fn)

    @overload
    def load(self: ImplementSample[CT], *collections: dict[Callable, None]) -> HarvestWrapper[CT]:
        ...

    @overload
    def load(
        self: Collectable[Concatenate[OverloadRecorder[Callable[P, R]], P1]], *collections: dict[Callable, None]
    ) -> HarvestWrapper[Callable[P, R]]:
        ...

    def load(self, *collections: dict[Callable, None]):  # type: ignore
        if not collections:
            raise TypeError("at least one collection is required")

        col = list(collections)
        init = col.pop(0)

        if not col:
            return HarvestWrapper(init)

        r = reduce(lambda x, y: {i: None for i in x if i in y}, col, init)
        return HarvestWrapper(r)


class HarvestWrapper(Generic[CT]):
    harvest: dict[CT, None]

    def __init__(self, harvest: dict[CT, None]):
        self.harvest = harvest

    @property
    def first(self) -> CT:
        if not self.harvest:
            raise NotImplementedError("cannot lookup any implementation with given arguments")

        return next(iter(self.harvest))

    @property
    def __call__(self):
        return self.first

    def __iter__(self):
        if not self.harvest:
            raise NotImplementedError("cannot lookup any implementation with given arguments")

        return iter(self.harvest)

    def __bool__(self):
        return bool(self.harvest)



from __future__ import annotations

import functools
from typing import Any, Callable

from .context import CollectContext
from .globals import COLLECTING_CONTEXT_VAR, GLOBAL_COLLECT_CONTEXT, GLOBAL_INSTANCE_CONTEXT, INSTANCE_CONTEXT_VAR
from .typing import TYPE_CHECKING, AssignKeeperCls, P, R, TEntity

if TYPE_CHECKING:
    from typing_extensions import Concatenate

    from .fn.base import Call, Fn
    from .fn.implement import FnImplementEntity, OverloadRecorder
    from .fn.record import FnImplement, FnRecord
    from .typing import CR, OutP


class scoped_collect(CollectContext):
    fn_implements: dict[FnImplement, FnRecord]
    _tocollect_list: dict[FnImplementEntity, None]
    finalize_cbs: list[Callable[[scoped_collect], Any]]
    cls: type | None = None

    def __init__(self) -> None:
        self.fn_implements = {}
        self.finalize_cbs = []
        self._tocollect_list = {}

    @classmethod
    def globals(cls):
        instance = cls()
        instance.fn_implements = GLOBAL_COLLECT_CONTEXT.fn_implements
        return instance

    @classmethod
    def env(cls):
        instance = cls()
        instance.fn_implements = COLLECTING_CONTEXT_VAR.get().fn_implements
        return instance

    def finalize(self):
        for cb in self.finalize_cbs:
            cb(self)

        for impl in self._tocollect_list:
            impl.collect(self)

    def on_collected(self, func: Callable[[scoped_collect], Any]):
        self.finalize_cbs.append(func)
        return func

    def remove_collected_callback(self, func: Callable[[scoped_collect], Any]):
        self.finalize_cbs.remove(func)
        return func

    @property
    def target(self):
        from .fn.implement import FnImplementEntity

        class LocalEndpoint:
            collector = self

            def __init_subclass__(cls, *, static: bool = False) -> None:
                self.cls = cls
                self.finalize()

                if static:
                    GLOBAL_INSTANCE_CONTEXT.instances[cls] = cls()

            @staticmethod
            def collect(entity: TEntity) -> TEntity:  # type: ignore
                return entity.collect(self)

            @staticmethod
            def ensure_self(func: Callable[Concatenate[Any, P], R]) -> Callable[P, R]:
                @functools.wraps(func)
                def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                    assert self.cls is not None

                    instance = INSTANCE_CONTEXT_VAR.get().instances[self.cls]
                    return func(instance, *args, **kwargs)

                return wrapper

            @staticmethod
            def impl(
                fn: Fn[Callable[Concatenate[OverloadRecorder[CR], OutP], Any], Any], *args: OutP.args, **kwargs: OutP.kwargs
            ) -> AssignKeeperCls[Call[..., CR]]:
                def inner(impl: Callable[Concatenate[Any, P], R] | FnImplementEntity[Callable[P, R]]):
                    if not isinstance(impl, FnImplementEntity):
                        if not hasattr(impl, "__wrapped__"):
                            impl_call = LocalEndpoint.ensure_self(impl)
                        else:
                            impl_call: Callable[P, R] = impl  # type: ignore

                        impl = FnImplementEntity(impl_call)

                    impl.add_target(fn, *args, **kwargs)
                    self._tocollect_list[impl] = None
                    return impl

                return inner  # type: ignore

        return LocalEndpoint



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypedDict, final

from .scoped import scoped_collect

if TYPE_CHECKING:
    from .context import CollectContext


class EntityAssignInfo(TypedDict, total=False):
    cls: type
    name: str
    annotation: Any


class BaseEntity:
    collect_context: CollectContext | None = None

    def collect(self, collector: CollectContext):
        self.collect_context = collector

        if isinstance(self.collect_context, scoped_collect):
            self.collect_context.on_collected(self._fallback_collected_callback)

        return self

    def assign_callback(self, info: EntityAssignInfo):
        ...

    @final
    def _fallback_collected_callback(self, collector):
        self.assign_callback({})

    def __set_name__(self, owner: type, name: str):
        if self.collect_context is None:
            return

        if isinstance(self.collect_context, scoped_collect):
            if owner is not self.collect_context.cls:
                return

            try:
                self.collect_context.remove_collected_callback(self._fallback_collected_callback)
            except ValueError:
                pass

            @self.collect_context.on_collected
            def _(collector: scoped_collect):
                if collector.cls is None:
                    return

                if name in collector.cls.__annotations__:
                    self.assign_callback({"cls": collector.cls, "name": name, "annotation": collector.cls.__annotations__[name]})
                else:
                    self.assign_callback({"cls": collector.cls, "name": name})



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, Protocol, Type, TypeVar

from typing_extensions import Concatenate

from ..entity import BaseEntity
from ..globals import iter_layout
from ..typing import CR, AssignKeeper, Call, OutP, P, R
from .compose import FnCompose
from .implement import FnImplementEntity, OverloadRecorder

if TYPE_CHECKING:
    from .record import FnRecord

K = TypeVar("K")

CCall = TypeVar("CCall", bound=Callable, covariant=True)
CCollect = TypeVar("CCollect", bound=Callable, covariant=True)
C = TypeVar("C", bound=Callable)


class ComposeShape(Protocol[CCollect, CCall]):
    @property
    def collect(self) -> CCollect:
        ...

    @property
    def call(self) -> CCall:
        ...


FnDef = Type[ComposeShape[CCollect, Callable[Concatenate["FnRecord", OutP], R]]]


class Fn(Generic[CCollect, CCall], BaseEntity):
    desc: FnCompose

    def __init__(self, compose: type[FnCompose]):
        self.desc = compose(self)

    @classmethod
    def declare(cls, desc: FnDef[C, OutP, R]) -> Fn[C, Callable[OutP, R]]:
        return cls(desc)  # type: ignore

    @property
    def impl(
        self: Fn[Callable[Concatenate[OverloadRecorder[CR], OutP], Any], Any],
    ) -> Callable[OutP, AssignKeeper[Call[..., CR]]]:
        def wrapper(*args: OutP.args, **kwargs: OutP.kwargs):
            def inner(impl: Callable[P, R] | FnImplementEntity[Callable[P, R]]):
                if not isinstance(impl, FnImplementEntity):
                    impl = FnImplementEntity(impl)

                impl.add_target(self, *args, **kwargs)
                return impl

            return inner

        return wrapper  # type: ignore

    def _call(self, *args, **kwargs):
        signature = self.desc.signature()

        for context in iter_layout(signature):
            if signature not in context.fn_implements:
                continue

            record = context.fn_implements[signature]
            return record.spec.desc.call(record, *args, **kwargs)
        else:
            raise NotImplementedError("cannot find any record with given fn declaration")

    @property
    def __call__(self) -> CCall:
        return self._call  # type: ignore



from __future__ import annotations

from typing import Any, Generic, TypeVar, overload

from typing_extensions import Self

from .globals import INSTANCE_CONTEXT_VAR

T = TypeVar("T")


class InstanceOf(Generic[T]):
    target: type[T]

    def __init__(self, target: type[T]) -> None:
        self.target = target

    @overload
    def __get__(self, instance: None, owner: type) -> Self:
        ...

    @overload
    def __get__(self, instance: Any, owner: type) -> T:
        ...

    def __get__(self, instance: Any, owner: type):
        if instance is None:
            return self

        return INSTANCE_CONTEXT_VAR.get().instances[self.target]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from .base import Fn
    from .overload import FnOverload


@dataclass(eq=True, frozen=True)
class FnRecord:
    spec: Fn
    scopes: dict[str, dict[Any, Any]] = field(default_factory=dict)
    entities: dict[frozenset[tuple[str, "FnOverload", Any]], Callable] = field(default_factory=dict)


@dataclass(eq=True, frozen=True)
class FnImplement:
    fn: Fn



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Generic

from typing_extensions import Concatenate

from ..context import CollectContext
from ..entity import BaseEntity
from ..typing import CR, P
from .record import FnRecord

if TYPE_CHECKING:
    from .base import Fn
    from .overload import FnOverload, TCollectValue


@dataclass
class OverloadRecorder(Generic[CR]):
    target: FnRecord
    implement: CR
    operators: list[tuple[str, FnOverload, Any]] = field(default_factory=list)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.done()

    def done(self):
        for name, rule, value in self.operators:
            rule.lay(self.target, value, self.implement, name=name)

        self.target.entities[frozenset(self.operators)] = self.implement

    def use(self, target: FnOverload[Any, TCollectValue, Any], value: TCollectValue, *, name: str | None = None):
        self.operators.append((name or target.name, target, value))
        return self


class FnImplementEntity(Generic[CR], BaseEntity):
    targets: list[tuple[Fn, tuple[Any, ...], dict[str, Any]]]
    impl: CR

    def __init__(self, impl: CR):
        self.targets = []
        self.impl = impl

    def add_target(self, fn: Fn[Callable[Concatenate[Any, P], Any], Any], *args: P.args, **kwargs: P.kwargs):
        self.targets.append((fn, args, kwargs))

    def collect(self, collector: CollectContext):
        super().collect(collector)

        for fn, args, kwargs in self.targets:
            record_signature = fn.desc.signature()
            if record_signature in collector.fn_implements:
                record = collector.fn_implements[record_signature]
            else:
                record = collector.fn_implements[record_signature] = FnRecord(fn)

            with OverloadRecorder(record, self.impl) as recorder:
                fn.desc.collect(recorder, *args, **kwargs)

        return self

    def _call(self, *args, **kwargs):
        return self.impl(*args, **kwargs)

    @property
    def __call__(self) -> CR:
        return self._call  # type: ignore

    @property
    def super(self) -> CR:
        return self.targets[0][0].__call__



from __future__ import annotations

import functools
from collections import defaultdict
from contextvars import ContextVar, copy_context
from typing import Any, Callable, Generator, Tuple

from .context import CollectContext, InstanceContext
from .typing import P, R, TEntity

GLOBAL_COLLECT_CONTEXT = CollectContext()
GLOBAL_INSTANCE_CONTEXT = InstanceContext()

COLLECTING_CONTEXT_VAR = ContextVar("CollectingContext", default=GLOBAL_COLLECT_CONTEXT)
LOOKUP_LAYOUT_VAR = ContextVar[Tuple[CollectContext, ...]]("LookupContext", default=(GLOBAL_COLLECT_CONTEXT,))
INSTANCE_CONTEXT_VAR = ContextVar("InstanceContext", default=GLOBAL_INSTANCE_CONTEXT)

ITER_BUCKET_VAR: ContextVar[defaultdict[Any, list[int]]] = ContextVar("LAYOUT_ITER_COLLECTIONS")


def _standalone_context(func: Callable[P, R]) -> Callable[P, R]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return copy_context().run(func, *args, **kwargs)

    return wrapper


@_standalone_context
def iter_layout(session_id: Any | None = None) -> Generator[CollectContext, None, None]:
    bucket = ITER_BUCKET_VAR.get(None)
    if bucket is None:
        bucket = defaultdict(lambda: [-1])
        ITER_BUCKET_VAR.set(bucket)

    stack = bucket[session_id]

    contexts = LOOKUP_LAYOUT_VAR.get()
    index = stack[-1]
    stack.append(index)

    try:
        for content in contexts[index + 1 :]:
            stack[-1] += 1
            yield content
    finally:
        stack.pop()
        if not stack:
            bucket.pop(session_id, None)


def global_collect(entity: TEntity) -> TEntity:
    return GLOBAL_COLLECT_CONTEXT.collect(entity)


def local_collect(entity: TEntity) -> TEntity:
    return COLLECTING_CONTEXT_VAR.get().collect(entity)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar

from typing_extensions import Concatenate, ParamSpec

if TYPE_CHECKING:
    from .entity import BaseEntity
    from .fn.implement import FnImplementEntity


R = TypeVar("R", covariant=True)
R1 = TypeVar("R1", covariant=True)
P = ParamSpec("P")
P1 = ParamSpec("P1")
InP = ParamSpec("InP")
OutP = ParamSpec("OutP")
CR = TypeVar("CR", covariant=True, bound=Callable)
CT = TypeVar("CT", bound=Callable)
TEntity = TypeVar("TEntity", bound="BaseEntity")


class Collectable(Protocol[InP]):
    def collect(self, *args: InP.args, **kwargs: InP.kwargs) -> Any:
        ...


class Call(Protocol[P, R]):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        ...


class AssignKeeper(Protocol[R]):
    def __call__(
        self: AssignKeeper[Call[..., Callable[P1, R1]]],
        implement: Callable[P1, R1] | FnImplementEntity[Callable[P1, R1]],
    ) -> FnImplementEntity[Callable[P1, R1]]:
        ...


class AssignKeeperCls(Protocol[R]):
    def __call__(
        self: AssignKeeperCls[Call[..., Callable[P1, R1]]],
        implement: Callable[Concatenate[Any, P1], R1] | FnImplementEntity[Callable[P1, R1]],
    ) -> FnImplementEntity[Callable[P1, R1]]:
        ...


class ImplementSample(Protocol[CR]):
    @property
    def implement_sample(self) -> CR:
        ...



from .base import Fn as Fn
from .compose import FnCompose as FnCompose
from .implement import FnImplementEntity as FnImplementEntity
from .implement import OverloadRecorder as OverloadRecorder
from .overload import FnOverload as FnOverload
from .record import FnImplement as FnImplement
from .record import FnRecord as FnRecord



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generic, TypeVar

from typing_extensions import final

if TYPE_CHECKING:
    from .record import FnRecord

TOverload = TypeVar("TOverload", bound="FnOverload", covariant=True)
TCallValue = TypeVar("TCallValue")
TCollectValue = TypeVar("TCollectValue")
TSignature = TypeVar("TSignature")


class FnOverload(Generic[TSignature, TCollectValue, TCallValue]):
    def __init__(self, name: str) -> None:
        self.name = name

    @final
    def dig(self, record: FnRecord, call_value: TCallValue, *, name: str | None = None) -> dict[Callable, None]:
        name = name or self.name
        if name not in record.scopes:
            raise NotImplementedError("cannot lookup any implementation with given arguments")

        return self.harvest(record.scopes[name], call_value)

    @final
    def lay(self, record: FnRecord, collect_value: TCollectValue, implement: Callable, *, name: str | None = None):
        name = name or self.name
        if name not in record.scopes:
            record.scopes[name] = {}

        collection = self.collect(record.scopes[name], self.digest(collect_value))
        collection[implement] = None

    def digest(self, collect_value: TCollectValue) -> TSignature:
        raise NotImplementedError

    def collect(self, scope: dict, signature: TSignature) -> dict[Callable, None]:
        raise NotImplementedError

    def harvest(self, scope: dict, call_value: TCallValue) -> dict[Callable, None]:
        raise NotImplementedError

    def access(self, scope: dict, signature: TSignature) -> dict[Callable, None] | None:
        raise NotImplementedError
